{-|
A parsing module, adapted from
\"Haskell: The Craft of Functional Programming\", 2nd ed., S. Thompson.
-}
module Parser where


import Data.Char (isLetter, isDigit)

{-|
Un parser este o functia care, dat un input (o lista de elemente de tip tchr), extrage o serie de tokeni (de tip ttok) care se potrivesc cu expresia. Daca doar o parte de la inceputul expresiei se potriveste, restul inputului va fi lasat neschimbat in a doua parte a perechii din rezultat.

Pot exista mai multe parsari posibile pentru acelasi input, deci outputul parserului este o lista de posibile parsari, fiecare parsare fiind o pereche din lista de tokeni extrasi si inputul ramas neprelucrat.

Inputul unui parser este format din entitati de tip tchr
-}
type Parser tchr ttok = [tchr] -> [([ttok], [tchr])]


{-
Vrem sa obtinem un parser care sa functioneze, de exemplu, pe expresii de forma:

NumExpr = Number | NumExpr.Op.Numxpr | (.NumExpr.)

-}


{-|
Parsarea unui singur caracter/token, care apartine unei multimi definite printr-o relatie de apartenenta (functia data in argumentul inSet).

Exemple:

parserSetToken isDigit "5*(2+33)" -> 5
parserSetToken isLetter "abaaaabbaaab" -> a
-}
parserSetToken :: (a -> Bool) -> Parser a a
parserSetToken _ [] = []
parserSetToken inSet (h:t)
	| inSet h	= [([h], t)]
	| otherwise	= []


{-|
Parsarea unui singur caracter/token, dat explicit (e.g. caractere speciale).

Exemplu:

parserSpecToken '(' "(2+3)" -> '('
-}
parserSpecToken :: Eq a => a -> Parser a a
parserSpecToken _ [] = []
parserSpecToken token (h:t)
	| h == token	= [([h], t)]
	| otherwise	= []


{-|
Concatenarea a doua parsere: pentru fiecare solutie (parsare) oferita de primul parser, al doilea parser preia inputul ramas si il parseaza. Solutia celui de-al doilea parser este asamblata cu solutia primului parser.

Concatenarea este definita aici ca operator infixat cu asociativitate dreapta si precedenta 5.

Atentie: in aceasta implementare, solutiile parserelor sunt concatenate, ceea ce reduce generalitatea implementarii.

Pentru o varianta de implementare unde solutiile sunt asamblate ca membri ai unei perechi, consultati implementarea lui Mihnea de la cursurile seriilor CA/CB.

Exemplu:

(bracketIn >.> star) "(*iviv" -> "(*"
-}
(>.>) :: Parser a b -> Parser a b -> Parser a b
infixr 5 >.>
(>.>) parser1 parser2 input = 
	[(result1 ++ result2, remainder)
	| (result1, remainder1) <- parser1 input
	, (result2, remainder) <- parser2 remainder1
	]


{-|
Alternarea a doua parsere: solutiile din lista intoarsa vor fi fie potriviri ale primului parser sau ale celui de-al doilea.

Alternarea este definita aici ca operator infixat cu asociativitate dreapta si precedenta 4 (mai mica decat a operatorului de concatenare.

Exemplu: 

(bracketIn >|> star) "(*iviv" -> "("
(bracketIn >|> star) "*(iviv" -> "*"
(plus >|> plus >.> plus) "++x" -> "+", "++"

-}
(>|>) :: Parser a b -> Parser a b -> Parser a b
infixr 4 >|>
(>|>) parser1 parser2 input = parser1 input ++ parser2 input

{-|
Consolidarea este operatiunea prin care rezultatul unui parser este transformat / prelucrat (eventual intr-un rezultat de alt tip)

Consolidarea este definita si ca operator mai jos.

Exemplu:

(plus >.> plus >$> (\x -> if x == "++" then "Inc" else x)) "++x" -> "Inc"
-}
consolidateParser :: Parser a b -> ([b] -> [c]) -> Parser a c
consolidateParser parser consolidate input =
	[	(consolidate result, remainder)
	|	(result, remainder) <- parser input
	]

{-|
Definirea consolidarii ca operator.
-}
infixr 3 >$>
(>$>) = consolidateParser


{-|
Extinderea unui parser pentru a potrivi un numar nelimitat de tokeni de acelasi fel.

Aici este varianta care ofera si consolidare. Daca nu se doreste consolidare, se poate modifica, sau se poate folosi functia id pentru consolidare.

Rezultate va fi o lista cu potriviri posibile, de lungimi crescatoare, inclusiv potrivirea nula.

Exemplu:

(extend digit id) "48451acbei" -> "48451"
-}
extend :: Parser a b -> ([b] -> [b]) -> Parser a b
extend parser consolidate =  (\i -> [([], i)]) >|> ((parser >.> extend parser consolidate) >$> consolidate)

{-|
Acelasi efec ca si extend, dar fara potrivirea de lungime nula.
-}
extend1 :: Parser a b -> ([b] -> [b]) -> Parser a b
extend1 parser consolidate =  consolidateParser (parser >.> extend parser consolidate) consolidate

{-|
Aplica un parser pe un input si intoarce un rezultat usor de citit, care contine cea mai lunga potrivire corecta.

Atentie! Aceasta implementare functioneaza numai pentru parseri cu output (tokeni) de tip [Char]

Exemplu:

parse numExpr "55*(2+6)+4" -> succes
parse numExpr "55*(2+veiqwe" -> succes pentru partea de expresie completa corecta - "55"
parse numExpr "55*(2+3)veiqwe" -> succes pentru partea de expresie completa corecta
parse numExpr "*3veiqwe" -> esec
-}
parse parser input
	|	result == []	= "failed"
	|	remainder == []	= "success: " ++ bestOutput
	|	otherwise		= "partial success: " ++ bestOutput ++ " | remaining: " ++ remainder
	where
		result = parser input
		(bestOutput, remainder) = last result



-- Aplicare pe expresii numerice cu numere, paranteze, + si *:
digit = parserSetToken isDigit
bracketIn = parserSpecToken '('
bracketOut = parserSpecToken ')'
star = parserSpecToken '*'
plus = parserSpecToken '+'
increment = plus >.> plus >$> consPlus
	where
		consPlus ['+', '+'] = "Inc_"

operator = plus >|> star >$> consOperator
	where
		consOperator _ = "Op"

number = extend1 digit consDigit
	where
		consDigit [d] = "N"
		consDigit (d : "N") = "N"
		
term = ((number >$> consNumExp) >|> (bracketIn >.> numExpr >.> bracketOut >$> consNumExp))
	where
		consNumExp "N" = "Exp"
		consNumExp "(Exp)" = "Exp"

numExpr = term >.> (extend (operator >.> term) consNumExp2) >$> consNumExp3
	where
		consNumExp2 "OpExpOpExp" = "OpExp"
		consNumExp2 "OpExp" = "OpExp"
		consNumExp3 "ExpOpExp" = "Exp"
		consNumExp3 "Exp" = "Exp"

